from __future__ import annotations
from anise import astro
import typing

@typing.final
class CelestialObjects:

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(cls):...
    EARTH: int = ...
    EARTH_MOON_BARYCENTER: int = ...
    JUPITER: int = ...
    JUPITER_BARYCENTER: int = ...
    MARS: int = ...
    MARS_BARYCENTER: int = ...
    MERCURY: int = ...
    MOON: int = ...
    NEPTUNE: int = ...
    NEPTUNE_BARYCENTER: int = ...
    PLUTO_BARYCENTER: int = ...
    SATURN: int = ...
    SATURN_BARYCENTER: int = ...
    SOLAR_SYSTEM_BARYCENTER: int = ...
    SUN: int = ...
    URANUS: int = ...
    URANUS_BARYCENTER: int = ...
    VENUS: int = ...

@typing.final
class Frames:

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(cls):...
    EARTH_ECLIPJ2000: astro.Frame = ...
    EARTH_ITRF93: astro.Frame = ...
    EARTH_J2000: astro.Frame = ...
    EARTH_MOON_BARYCENTER_J2000: astro.Frame = ...
    EME2000: astro.Frame = ...
    IAU_EARTH_FRAME: astro.Frame = ...
    IAU_JUPITER_FRAME: astro.Frame = ...
    IAU_MARS_FRAME: astro.Frame = ...
    IAU_MERCURY_FRAME: astro.Frame = ...
    IAU_MOON_FRAME: astro.Frame = ...
    IAU_NEPTUNE_FRAME: astro.Frame = ...
    IAU_SATURN_FRAME: astro.Frame = ...
    IAU_URANUS_FRAME: astro.Frame = ...
    IAU_VENUS_FRAME: astro.Frame = ...
    JUPITER_BARYCENTER_J2000: astro.Frame = ...
    MARS_BARYCENTER_J2000: astro.Frame = ...
    MERCURY_J2000: astro.Frame = ...
    MOON_J2000: astro.Frame = ...
    MOON_ME_DE421_FRAME: astro.Frame = ...
    MOON_ME_DE440_ME421_FRAME: astro.Frame = ...
    MOON_ME_FRAME: astro.Frame = ...
    MOON_PA_DE421_FRAME: astro.Frame = ...
    MOON_PA_DE440_FRAME: astro.Frame = ...
    MOON_PA_FRAME: astro.Frame = ...
    NEPTUNE_BARYCENTER_J2000: astro.Frame = ...
    PLUTO_BARYCENTER_J2000: astro.Frame = ...
    SATURN_BARYCENTER_J2000: astro.Frame = ...
    SSB_J2000: astro.Frame = ...
    SUN_J2000: astro.Frame = ...
    URANUS_BARYCENTER_J2000: astro.Frame = ...
    VENUS_J2000: astro.Frame = ...

@typing.final
class Orientations:

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(cls):...
    ECLIPJ2000: int = ...
    IAU_EARTH: int = ...
    IAU_JUPITER: int = ...
    IAU_MARS: int = ...
    IAU_MERCURY: int = ...
    IAU_MOON: int = ...
    IAU_NEPTUNE: int = ...
    IAU_SATURN: int = ...
    IAU_URANUS: int = ...
    IAU_VENUS: int = ...
    ITRF93: int = ...
    J2000: int = ...
    MOON_ME: int = ...
    MOON_ME_DE421: int = ...
    MOON_ME_DE440_ME421: int = ...
    MOON_PA: int = ...
    MOON_PA_DE421: int = ...
    MOON_PA_DE440: int = ...

@typing.final
class UsualConstants:

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        """Initialize self.  See help(type(self)) for accurate signature."""

    def __new__(cls):...
    MEAN_EARTH_ANGULAR_VELOCITY_DEG_S: float = ...
    MEAN_MOON_ANGULAR_VELOCITY_DEG_S: float = ...
    SPEED_OF_LIGHT_KM_S: float = ...